import click
from kservehelper.docker import build as build_image
from kservehelper.docker import push as push_image


def create_cli() -> click.Group:
    @click.group()
    def cli():
        pass

    @cli.command()
    @click.argument("folder", type=click.STRING)
    @click.option(
        "--quiet",
        type=click.BOOL,
        default=False,
        is_flag=True
    )
    def build(folder, quiet):
        build_image(folder, quiet)

    @cli.command()
    @click.argument("folder", type=click.STRING)
    def push(folder):
        push_image(folder)

    return cli


def main():
    _cli = create_cli()
    _cli()



import base64
import io
from PIL import Image, ImageFilter
from kservehelper import KServeModel
from kservehelper.types import Input, Path


class Model:

    def load(self):
        pass

    def predict(
            self,
            image: str = Input(
                description="Base64 encoded image",
                default=""
            ),
            radius: float = Input(
                description="Standard deviation of the Gaussian kernel",
                default=2
            )
    ) -> Path:
        if image == "":
            raise ValueError("The input image is not set")
        im_binary = base64.b64decode(image)
        input_image = Image.open(io.BytesIO(im_binary))
        output_image = input_image.filter(ImageFilter.GaussianBlur(radius))
        output_path = KServeModel.generate_filepath("image.jpg")
        output_image.save(output_path)
        return Path(output_path)


if __name__ == "__main__":
    KServeModel.serve("test-blur-image", Model)



import os
import json
import aiohttp
import asyncio
import aiofiles
import requests
import concurrent.futures
from typing import List
from kservehelper.types import Path


def upload_files(webhook_url: str, paths: List[Path], timeout=60):
    # outputs = asyncio.run(_upload_v2(webhook_url, paths, timeout))
    # return outputs
    outputs = _upload_multithread(webhook_url, paths, timeout)
    return {"output": outputs}


def _upload(webhook_url: str, paths: List[Path], timeout):
    outputs = []
    for path in paths:
        filepath = str(path)
        with open(filepath, "rb") as f:
            files = {"file": (filepath, f)}
            response = requests.post(webhook_url, files=files, timeout=timeout)
        if response.status_code != 200:
            raise RuntimeError("failed to call upload webhook")
        r = json.loads(response.text)
        outputs.append(r["url"])
    return outputs


def _upload_multithread(webhook_url: str, paths: List[Path], timeout):
    def _make_request(filepath):
        with open(filepath, "rb") as f:
            files = {"file": (filepath, f)}
            for i in range(3):
                try:
                    response = requests.post(webhook_url, files=files, timeout=timeout)
                    if response.status_code != 200:
                        raise RuntimeError(f"response status code is {response.status_code}")
                    r = json.loads(response.text)
                    return filepath, r["url"]
                except Exception as e:
                    print(f"ERROR: {e}")
                    print(f"Retrying {i + 1}...")
            return None

    with concurrent.futures.ThreadPoolExecutor(max_workers=4) as executor:
        jobs = [executor.submit(_make_request, str(path)) for path in paths]
        outputs = [job.result() for job in concurrent.futures.as_completed(jobs)]
        for path in paths:
            file = str(path)
            if os.path.exists(file):
                os.remove(file)
        if None in outputs:
            raise RuntimeError("failed to call upload webhook")

    filename2url = dict(outputs)
    return [filename2url[str(path)] for path in paths]


async def _async_upload_v1(webhook_url: str, paths: List[Path], timeout):
    outputs = []
    async with aiohttp.ClientSession() as session:
        for path in paths:
            filepath = str(path)
            with open(filepath, "rb") as f:
                data = aiohttp.FormData()
                data.add_field("file", f, filename=filepath)
                async with session.post(webhook_url, data=data, timeout=timeout) as response:
                    if not response.ok:
                        raise RuntimeError("failed to call upload webhook")
                    text = await response.text()
                    r = json.loads(text)
                    outputs.append(r["url"])
    return outputs


async def _async_upload_v2(webhook_url, paths, timeout):
    async def _upload_file(session, local_path):
        async with aiofiles.open(local_path, "rb") as f:
            data = aiohttp.FormData()
            data.add_field("file", f, filename=local_path)
            async with session.post(webhook_url, data=data, timeout=timeout) as response:
                if not response.ok:
                    raise RuntimeError("failed to call upload webhook")
                text = await response.text()
                r = json.loads(text)
                return local_path, r["url"]

    async with aiohttp.ClientSession() as sess:
        urls = await asyncio.gather(*[_upload_file(sess, str(path)) for path in paths])
        filename2url = dict(urls)
        return [filename2url[str(path)] for path in paths]



import os
import torch
from typing import List
from diffusers import DiffusionPipeline
from kservehelper import KServeModel
from kservehelper.types import Input, Path


# https://huggingface.co/docs/diffusers/v0.21.0/using-diffusers/sdxl


class StableDiffusionXL:

    def __init__(self):
        self.device = None
        self.base = None
        self.refiner = None
        self.use_refiner = os.getenv("USE_REFINER")
        self.use_torch_compile = os.getenv("USE_TORCH_COMPILE")

    def load(self):
        model_path = "/mnt/models/stable-diffusion-xl-base-1.0"
        refiner_path = "/mnt/models/stable-diffusion-xl-refiner-1.0"

        self.device = "cuda" if torch.cuda.is_available() else "cpu"
        self.base = DiffusionPipeline.from_pretrained(
            model_path,
            torch_dtype=torch.float16,
            variant="fp16",
            use_safetensors=True
        ).to(self.device)

        if self.use_torch_compile:
            self.base.unet = torch.compile(
                self.base.unet, mode="reduce-overhead", fullgraph=True
            )

        if self.use_refiner:
            self.refiner = DiffusionPipeline.from_pretrained(
                refiner_path,
                text_encoder_2=self.base.text_encoder_2,
                vae=self.base.vae,
                torch_dtype=torch.float16,
                use_safetensors=True,
                variant="fp16",
            ).to(self.device)

    def predict(
            self,
            prompt: str = Input(
                description="Input prompt",
                default="a dog playing on the street"
            ),
            guidance_scale: float = Input(
                description="Guidance scale",
                default=5.0
            ),
            height: int = Input(
                description="Image height",
                default=1024
            ),
            width: int = Input(
                description="Image width",
                default=1024
            ),
            num_inference_steps: int = Input(
                description="The number of inference steps",
                default=50
            ),
            num_images_per_prompt: int = Input(
                description="The number of images to generate per prompt",
                default=1
            ),
            seed: int = Input(
                description="Random seed",
                default=12345
            )
    ) -> List[Path]:
        generator = torch.Generator(self.device)
        generator.manual_seed(seed)

        if self.use_refiner:
            image = self.base(
                prompt=prompt,
                denoising_end=0.8,
                output_type="latent",
                guidance_scale=guidance_scale,
                height=height,
                width=width,
                num_inference_steps=num_inference_steps,
                num_images_per_prompt=num_images_per_prompt,
                generator=generator
            ).images
            images = self.refiner(
                prompt=prompt,
                denoising_start=0.8,
                image=image,
                guidance_scale=guidance_scale,
                num_inference_steps=num_inference_steps,
                num_images_per_prompt=num_images_per_prompt,
                generator=generator
            ).images
        else:
            images = self.base(
                prompt=prompt,
                guidance_scale=guidance_scale,
                height=height,
                width=width,
                num_inference_steps=num_inference_steps,
                num_images_per_prompt=num_images_per_prompt,
                generator=generator
            ).images

        output_paths = []
        for i, image in enumerate(images):
            output_path = KServeModel.generate_filepath(f"image_{i}.jpg")
            image.save(output_path)
            output_paths.append(Path(output_path))
        return output_paths


if __name__ == "__main__":
    KServeModel.serve("stable-diffusion-xl", StableDiffusionXL)



# Copyright 2022 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import asyncio
import concurrent.futures
import multiprocessing
import signal
import socket
from distutils.util import strtobool
from multiprocessing import Process
from typing import Dict, List, Optional, Union

from ray import serve as rayserve
from ray.serve.api import Deployment, RayServeHandle

from kserve.logging import KSERVE_LOG_CONFIG, logger
from kserve.model import Model
from kserve.model_repository import ModelRepository
from kserve.protocol.dataplane import DataPlane
from kserve.protocol.grpc.server import GRPCServer
from kserve.protocol.model_repository_extension import ModelRepositoryExtension
from kservehelper.kserve.rest.server import UvicornServer
from kserve.utils import utils

DEFAULT_HTTP_PORT = 8080
DEFAULT_GRPC_PORT = 8081

parser = argparse.ArgumentParser(add_help=False)
parser.add_argument("--http_port", default=DEFAULT_HTTP_PORT, type=int,
                    help="The HTTP Port listened to by the model server.")
parser.add_argument("--grpc_port", default=DEFAULT_GRPC_PORT, type=int,
                    help="The GRPC Port listened to by the model server.")
parser.add_argument("--workers", default=1, type=int,
                    help="The number of workers for multi-processing.")
parser.add_argument("--max_threads", default=4, type=int,
                    help="The number of max processing threads in each worker.")
parser.add_argument('--max_asyncio_workers', default=None, type=int,
                    help='Max number of asyncio workers to spawn')
parser.add_argument("--enable_grpc", default=True, type=lambda x: bool(strtobool(x)),
                    help="Enable gRPC for the model server")
parser.add_argument("--enable_docs_url", default=False, type=lambda x: bool(strtobool(x)),
                    help="Enable docs url '/docs' to display Swagger UI.")
parser.add_argument("--enable_latency_logging", default=True, type=lambda x: bool(strtobool(x)),
                    help="Output a log per request with latency metrics.")
parser.add_argument("--configure_logging", default=True, type=lambda x: bool(strtobool(x)),
                    help="Whether to configure KServe and Uvicorn logging")
parser.add_argument("--log_config_file", default=None, type=str,
                    help="File path containing UvicornServer's log config. Needs to be a yaml or json file.")
parser.add_argument("--access_log_format", default=None, type=str,
                    help="Format to set for the access log (provided by asgi-logger).")

args, _ = parser.parse_known_args()


class ModelServer:
    """KServe ModelServer

    Args:
        http_port (int): HTTP port. Default: ``8080``.
        grpc_port (int): GRPC port. Default: ``8081``.
        workers (int): Number of workers for uvicorn. Default: ``1``.
        max_threads (int): Max number of processing threads. Default: ``4``
        max_asyncio_workers (int): Max number of AsyncIO threads. Default: ``None``
        registered_models (ModelRepository): Model repository with registered models.
        enable_grpc (bool): Whether to turn on grpc server. Default: ``True``
        enable_docs_url (bool): Whether to turn on ``/docs`` Swagger UI. Default: ``False``.
        enable_latency_logging (bool): Whether to log latency metric. Default: ``True``.
        configure_logging (bool): Whether to configure KServe and Uvicorn logging. Default: ``True``.
        log_config (dict or str): File path or dict containing log config. Default: ``None``.
        access_log_format (string): Format to set for the access log (provided by asgi-logger). Default: ``None``
    """

    def __init__(self, http_port: int = args.http_port,
                 grpc_port: int = args.grpc_port,
                 workers: int = args.workers,
                 max_threads: int = args.max_threads,
                 max_asyncio_workers: int = args.max_asyncio_workers,
                 registered_models: ModelRepository = ModelRepository(),
                 enable_grpc: bool = args.enable_grpc,
                 enable_docs_url: bool = args.enable_docs_url,
                 enable_latency_logging: bool = args.enable_latency_logging,
                 configure_logging: bool = args.configure_logging,
                 log_config: Optional[Union[Dict, str]] = args.log_config_file,
                 access_log_format: str = args.access_log_format):
        self.registered_models = registered_models
        self.http_port = http_port
        self.grpc_port = grpc_port
        self.workers = workers
        self.max_threads = max_threads
        self.max_asyncio_workers = max_asyncio_workers
        self.enable_grpc = enable_grpc
        self.enable_docs_url = enable_docs_url
        self.enable_latency_logging = enable_latency_logging
        self.dataplane = DataPlane(model_registry=registered_models)
        self.model_repository_extension = ModelRepositoryExtension(
            model_registry=self.registered_models)
        self._grpc_server = None
        if self.enable_grpc:
            self._grpc_server = GRPCServer(grpc_port, self.dataplane,
                                           self.model_repository_extension)
        # Logs can be passed as a path to a file or a dictConfig.
        # We rely on Uvicorn to configure the loggers for us.
        if configure_logging:
            self.log_config = log_config if log_config is not None else KSERVE_LOG_CONFIG
        else:
            # By setting log_config to None we tell Uvicorn not to configure logging
            self.log_config = None

        self.access_log_format = access_log_format

    def start(self, models: Union[List[Model], Dict[str, Deployment]]) -> None:
        if isinstance(models, list):
            for model in models:
                if isinstance(model, Model):
                    self.register_model(model)
                    # pass whether to log request latency into the model
                    model.enable_latency_logging = self.enable_latency_logging
                else:
                    raise RuntimeError("Model type should be 'Model'")
        elif isinstance(models, dict):
            if all([isinstance(v, Deployment) for v in models.values()]):
                # TODO: make this port number a variable
                rayserve.start(detached=True, http_options={"host": "0.0.0.0", "port": 9071})
                for key in models:
                    models[key].deploy()
                    handle = models[key].get_handle()
                    self.register_model_handle(key, handle)
            else:
                raise RuntimeError("Model type should be RayServe Deployment")
        else:
            raise RuntimeError("Unknown model collection types")

        if self.max_asyncio_workers is None:
            # formula as suggest in https://bugs.python.org/issue35279
            self.max_asyncio_workers = min(32, utils.cpu_count() + 4)
        logger.info(f"Setting max asyncio worker threads as {self.max_asyncio_workers}")
        asyncio.get_event_loop().set_default_executor(
            concurrent.futures.ThreadPoolExecutor(max_workers=self.max_asyncio_workers))

        async def serve():
            logger.info(f"Starting uvicorn with {self.workers} workers")
            loop = asyncio.get_event_loop()
            for sig in [signal.SIGINT, signal.SIGTERM, signal.SIGQUIT]:
                loop.add_signal_handler(
                    sig, lambda s=sig: asyncio.create_task(self.stop(sig=s))
                )
            if self.workers == 1:
                self._rest_server = UvicornServer(self.http_port, [],
                                                  self.dataplane, self.model_repository_extension,
                                                  self.enable_docs_url,
                                                  log_config=self.log_config,
                                                  access_log_format=self.access_log_format)
                await self._rest_server.run()
            else:
                # Since py38 MacOS/Windows defaults to use spawn for starting multiprocessing.
                # https://docs.python.org/3/library/multiprocessing.html#contexts-and-start-methods
                # Spawn does not work with FastAPI/uvicorn in multiprocessing mode, use fork for multiprocessing
                # https://github.com/tiangolo/fastapi/issues/1586
                serversocket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                serversocket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                serversocket.bind(('0.0.0.0', self.http_port))
                serversocket.listen(5)
                multiprocessing.set_start_method('fork')
                server = UvicornServer(self.http_port, [serversocket],
                                       self.dataplane, self.model_repository_extension,
                                       self.enable_docs_url, log_config=self.log_config,
                                       access_log_format=self.access_log_format)
                for _ in range(self.workers):
                    p = Process(target=server.run_sync)
                    p.start()

        async def servers_task():
            servers = [serve()]
            if self.enable_grpc:
                servers.append(self._grpc_server.start(self.max_threads))
            await asyncio.gather(*servers)

        asyncio.run(servers_task())

    async def stop(self, sig: Optional[int] = None):
        logger.info("Stopping the model server")
        if self._rest_server:
            logger.info("Stopping the rest server")
            await self._rest_server.stop()
        if self._grpc_server:
            logger.info("Stopping the grpc server")
            await self._grpc_server.stop(sig)

    def register_model_handle(self, name: str, model_handle: RayServeHandle):
        self.registered_models.update_handle(name, model_handle)
        logger.info("Registering model handle: %s", name)

    def register_model(self, model: Model):
        if not model.name:
            raise Exception(
                "Failed to register model, model.name must be provided.")
        self.registered_models.update(model)
        logger.info("Registering model: %s", model.name)



import unittest
from typing import Dict
from kservehelper.model import KServeModel
from kservehelper.types import Input


class CustomTransform:

    def preprocess(
            self,
            prompt: str = Input(
                description="Input prompt",
                default="standing, (full body)++"
            )
    ) -> Dict:
        return {"outputs": prompt.upper()}

    def postprocess(self, infer_response: Dict) -> Dict:
        outputs = infer_response["outputs"] + " ABC"
        return {"outputs": outputs}


class TestKServeModel(unittest.TestCase):

    def test_tansform(self):
        payload = {"prompt": "test test", "upload_webhook": "http://localhost"}
        model = KServeModel("test", CustomTransform)
        outputs = model.postprocess(model.preprocess(payload))
        self.assertEqual(outputs["outputs"], "TEST TEST ABC")


if __name__ == "__main__":
    unittest.main()



# Copyright 2022 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Optional, Union, Dict, List

from fastapi import Request, Response

from kserve.errors import ModelNotReady
from kserve.protocol.dataplane import DataPlane
from kserve.protocol.model_repository_extension import ModelRepositoryExtension


class V1Endpoints:
    """KServe V1 Endpoints
    """

    def __init__(self, dataplane: DataPlane, model_repository_extension: Optional[ModelRepositoryExtension] = None):
        self.model_repository_extension = model_repository_extension
        self.dataplane = dataplane

    async def models(self) -> Dict[str, List[str]]:
        """Get a list of models in the model registry.

        Returns:
            Dict[str, List[str]]: List of model names.
        """
        return {"models": list(self.dataplane.model_registry.get_models().keys())}

    async def model_ready(self, model_name: str) -> Dict[str, Union[str, bool]]:
        """Check if a given model is ready.

        Args:
            model_name (str): Model name.

        Returns:
            Dict[str, Union[str, bool]]: Name of the model and whether it's ready.
        """
        model_ready = self.dataplane.model_ready(model_name)

        if not model_ready:
            raise ModelNotReady(model_name)

        return {"name": model_name, "ready": model_ready}

    async def predict(self, model_name: str, request: Request) -> Union[Response, Dict]:
        """Predict request handler.

        It sends the request to the dataplane where the model will process the request body.

        Args:
            model_name (str): Model name.
            request (Request): Raw request object.

        Returns:
            Dict|Response: Model inference response.
        """
        body = await request.body()
        headers = dict(request.headers.items())
        response, response_headers = await self.dataplane.infer(model_name=model_name, body=body, headers=headers)

        if not isinstance(response, dict):
            return Response(content=response, headers=response_headers)
        return response

    async def explain(self, model_name: str, request: Request) -> Union[Response, Dict]:
        """Explain handler.

        Args:
            model_name (str): Model name.
            request (Request): Raw request object.

        Returns:
            Dict: Explainer output.
        """
        body = await request.body()
        headers = dict(request.headers.items())
        response, response_headers = await self.dataplane.explain(model_name=model_name, body=body, headers=headers)

        if not isinstance(response, dict):
            return Response(content=response, headers=response_headers)
        return response

    async def docs(self, model_name: str) -> Dict:
        model = self.dataplane.model_registry.get_model(model_name)
        if hasattr(model, "docs"):
            return model.docs()
        else:
            return {}
